@@ -1,28 +1,48 @@
 """
 Implements common methods for any workflow manager
 """
+import abc
 import time
 import logging
 from uuid import uuid4
-import abc
+from argparse import Namespace
 from collections import defaultdict
+from typing import Optional, Generator, Protocol, List, Union, Dict
 
-from .script import BaseScript
+from .script import BaseScript, JobKey, JobDict
 
 
 logger = logging.getLogger(__name__)
 logger.setLevel(logging.INFO)
 
 
-class CachedFinishedJobsMixin:
+class WorkFlowManagerProtocol(Protocol):
+    @abc.abstractmethod
+    def get_owned_jobs(self, project_id: Optional[int] = None, **kwargs) -> Generator[JobDict, None, None]:
+        ...
+
+    @abc.abstractmethod
+    def get_finished_owned_jobs(self, project_id: Optional[int] = None, **kwargs) -> Generator[JobDict, None, None]:
+        ...
+
+    @abc.abstractmethod
+    def workflow_loop(self) -> bool:
+        """Implement here your loop code. Return True if want to continue looping,
+        False for immediate stop of the job. For continuous looping you also need
+        to set `loop_mode`
+        """
+        ...
+
+
+class CachedFinishedJobsMixin(WorkFlowManagerProtocol):
     def __init__(self):
         super().__init__()
-        self.__finished_cache = {}
-        self.__update_finished_cache_called = defaultdict(bool)
+        self.__finished_cache: Dict[JobKey, str] = {}
+        self.__update_finished_cache_called: Dict[int, bool] = defaultdict(bool)
 
-    def update_finished_cache(self, project_id):
-        logger.debug("Initiating finished cache update.")
+    def update_finished_cache(self, project_id: int):
         if not self.__update_finished_cache_called[project_id]:
+            logger.info("Initiating finished cache update.")
             for job in self.get_owned_jobs(project_id, state=["finished"], meta=["close_reason"]):
                 if job["key"] in self.__finished_cache:
                     break
@@ -30,8 +50,23 @@ class CachedFinishedJobsMixin:
             logger.info(f"Finished jobs cache length: {len(self.__finished_cache)}")
         self.__update_finished_cache_called[project_id] = True
 
-    def is_finished(self, jobkey):
-        project_id = jobkey.split("/", 1)[0]
+    def get_finished_owned_jobs(self, project_id: Optional[int] = None, **kwargs) -> Generator[JobDict, None, None]:
+        kwargs.setdefault("meta", []).append("close_reason")
+        # only use for updating finished cache if no filter/limit imposed.
+        update_finished_cache = not set(kwargs.keys()).intersection(["count", "has_tag", "lacks_tag"])
+        finished_cache = []
+        for job in super().get_finished_owned_jobs(project_id, **kwargs):
+            if update_finished_cache:
+                finished_cache.append((job["key"], job["close_reason"]))
+            yield job
+        logger.info(f"Preread {len(finished_cache)} finished jobs.")
+        while finished_cache:
+            key, close_reason = finished_cache.pop()
+            self.__finished_cache[key] = close_reason
+        logger.info(f"Finished jobs cache length: {len(self.__finished_cache)}")
+
+    def is_finished(self, jobkey: JobKey) -> Optional[str]:
+        project_id = int(jobkey.split("/", 1)[0])
         self.update_finished_cache(project_id)
         return self.__finished_cache.get(jobkey)
 
@@ -40,7 +75,7 @@ class CachedFinishedJobsMixin:
             self.__update_finished_cache_called[project_id] = False
 
 
-class WorkFlowManager(BaseScript, abc.ABC):
+class WorkFlowManager(BaseScript, WorkFlowManagerProtocol):
 
     # --max-running-job command line option overrides it
     default_max_jobs = float("inf")
@@ -72,7 +107,7 @@ class WorkFlowManager(BaseScript, abc.ABC):
         if not self.name:
             self.argparser.error("Manager name not set.")
 
-    def get_owned_jobs(self, project_id=None, **kwargs):
+    def get_owned_jobs(self, project_id: Optional[int] = None, **kwargs) -> Generator[JobDict, None, None]:
         assert self.flow_id, "This job doesn't have a flow id."
         assert self.name, "This job doesn't have a name."
         assert "has_tag" not in kwargs, "Filtering by flow id requires no extra has_tag."
@@ -88,11 +123,11 @@ class WorkFlowManager(BaseScript, abc.ABC):
                 yield job
 
     @staticmethod
-    def generate_flow_id():
+    def generate_flow_id() -> str:
         return str(uuid4())
 
     @property
-    def max_running_jobs(self):
+    def max_running_jobs(self) -> int:
         return self.args.max_running_jobs
 
     def add_argparser_options(self):
@@ -115,13 +150,19 @@ class WorkFlowManager(BaseScript, abc.ABC):
                                     Default: %(default)s",
         )
 
-    def parse_args(self):
+    def parse_args(self) -> Namespace:
         args = super().parse_args()
         if not self.name:
             self.name = args.name
         return args
 
-    def wait_for(self, jobs_keys, interval=60, timeout=float("inf"), heartbeat=None):
+    def wait_for(
+        self,
+        jobs_keys: Union[JobKey, List[JobKey]],
+        interval: int = 60,
+        timeout: float = float("inf"),
+        heartbeat: Optional[int] = None,
+    ):
         """Waits until all given jobs are not running anymore or until the
         timeout is reached, if a heartbeat is given it'll log an entry every
         heartbeat seconds (considering the interval), otherwise it'll log an
@@ -161,10 +202,32 @@ class WorkFlowManager(BaseScript, abc.ABC):
             self.add_job_tags(tags=inherited_tags)
             self.is_resumed = True
 
+    def get_finished_owned_jobs(self, project_id: Optional[int] = None, **kwargs) -> Generator[JobDict, None, None]:
+        kwargs.pop("state", None)
+        for job in self.get_owned_jobs(project_id, state=["finished"], **kwargs):
+            yield job
+
     def resume_workflow(self):
-        """
-        Implement resume logic
-        """
+        rcount = 0
+        for job in self.get_owned_jobs(state=["running", "pending"], meta=["spider_args", "job_cmd", "tags", "spider"]):
+            self.resume_running_job_hook(job)
+            logger.info(f"Found running job {job['key']}")
+            rcount += 1
+        if rcount > 0:
+            logger.info(f"Found a total of {rcount} running children jobs.")
+
+        fcount = 0
+        logger.info("Searching finished children jobs...")
+        for job in self.get_finished_owned_jobs(meta=["spider_args", "job_cmd", "tags", "spider"]):
+            self.resume_finished_job_hook(job)
+            fcount += 1
+        if fcount > 0:
+            logger.info(f"Found a total of {fcount} finished children jobs.")
+
+    def resume_running_job_hook(self, job: JobDict):
+        pass
+
+    def resume_finished_job_hook(self, job: JobDict):
         pass
 
     def __on_start(self):
@@ -174,14 +237,6 @@ class WorkFlowManager(BaseScript, abc.ABC):
         self.on_start()
         self.workflow_loop_enabled = True
 
-    @abc.abstractmethod
-    def workflow_loop(self):
-        """Implement here your loop code. Return True if want to continue looping,
-        False for immediate stop of the job. For continuous looping you also need
-        to set `loop_mode`
-        """
-        pass
-
     def on_close(self):
         pass
 
